; Calculating Bezier surfaces;
; Implemented in Flat assembler by Maciej Guba,
; http://macgub.co.pl
; If site not works try search the web for keyword 'macgub'.
; Requires SSE3



; I use this equation:
; x(s,t)=(1-s)^3*(x11*a+x12*b+x13*c+x14*d)+
;	 +3*(1-s)^2*s*(x21*a+x22*b+x23*c+x24*d)+
;	 +3*(1-s)*s^2*(x31*a+x32*b+x33*c+x34*d)+
;	 +s^3*(x41*a+x42*b+x43*c+x44*d)
; y(s,t) and z(s,t) similar...
;  a = (1-t)^3
;  b = 3*(1-t)^2*t
;  c = 3*(1-t)*t^2
;  d = t^3




calc_bezier_surface:
; calc single surface, write vertices to [edi],
; faces to [ebx]
;  in  :
;	esi - ptr  -> x11, y11, z11 .... x44, y44, z44 ; dword
;	edx - quality -> s shl 16 t
;	edi - ptr to area of calculated vertices
;	ebx - ptr to area of calculated faces (triangles)
;  out :
;	ecx - vertices count
;	edx - triangles count
;	edi - ptr to end of calculated vertices
;	ebx - ptr to end of calculated faces

	push	 ebp
	mov	 ebp,esp
	sub	 esp,290
	and	ebp,-16

  .t	       equ [ebp-4]
  .s	       equ [ebp-8]
  .ct	       equ dword [ebp-12] ; integer counter t
  .cs	       equ dword [ebp-16] ; int counter s
  .vert_countB equ dword [ebp-20]
  .tri_countB  equ dword [ebp-24]

  .x1 equ ebp-48	;      .x41-x44
  .x2 equ ebp-64	;      .x31-x34
  .x3 equ ebp-80	;      .x21-x24
  .x4 equ ebp-96	;      .x11-x14

  .y1 equ ebp-112	;      .y41-y44
  .y2 equ ebp-128	;      .y31-y34
  .y3 equ ebp-144	;      .y21-y24
  .y4 equ ebp-160	;      .y11-y14

  .z1 equ ebp-176	;      .z41-z44
  .z2 equ ebp-192	;      .z31-z34
  .z3 equ ebp-208	;      .z21-z24
  .z4 equ ebp-224	;      .z11-z14

  .ds equ  [ebp-228]	; delta s
  .dt equ  [ebp-232]	; delta t
  .d  equ  [ebp-244]
  .c  equ  [ebp-248]
  .b  equ  [ebp-252]
  .a  equ  [ebp-256]



   push  dword[esi+36]
   pop	 dword[.x1+12]
   push  dword[esi+24]
   pop	 dword[.x1+8]
   push  dword[esi+12]
   pop	 dword[.x1+4]
   push  dword[esi]
   pop	 dword[.x1]

 ; first convert derives for easier calculating.


   push dword[esi+36+12]    ; shuffle x
   pop	dword[.x2]
   push dword[esi+36+12*2]
   pop	dword[.x2+4]
   push dword[esi+36+12*3]
   pop	dword[.x2+8]
   push dword[esi+36+12*4]
   pop	dword[.x2+12]

   push dword[esi+36+12*5]
   pop	dword[.x3]
   push dword[esi+36+12*6]
   pop	dword[.x3+4]
   push dword[esi+36+12*7]
   pop	dword[.x3+8]
   push dword[esi+36+12*8]
   pop	dword[.x3+12]

   push dword[esi+36+12*9]
   pop	dword[.x4]
   push dword[esi+36+12*10]
   pop	dword[.x4+4]
   push dword[esi+36+12*11]
   pop	dword[.x4+8]
   push dword[esi+36+12*12]
   pop	dword[.x4+12]
   

   push dword[esi+4]	      ;shuffle y
   pop	dword[.y1]
   push dword[esi+12+4]
   pop	dword[.y1+4]
   push dword[esi+24+4]
   pop	dword[.y1+8]
   push dword[esi+36+4]
   pop	dword[.y1+12]

   push dword[esi+36+12+4]
   pop	dword[.y2]
   push dword[esi+36+12*2+4]
   pop	dword[.y2+4]
   push dword[esi+36+12*3+4]
   pop	dword[.y2+8]
   push dword[esi+36+12*4+4]
   pop	dword[.y2+12]

   lea	eax,[esi+36+12*5+4]

   push dword[eax]
   pop	dword[.y3]
   push dword[eax+12]
   pop	dword[.y3+4]
   push dword[eax+12*2]
   pop	dword[.y3+8]
   push dword[eax+12*3]
   pop	dword[.y3+12]

   push dword[eax+12*4]
   pop	dword[.y4]
   push dword[eax+12*5]
   pop	dword[.y4+4]
   push dword[eax+12*6]
   pop	dword[.y4+8]
   push dword[eax+12*7]
   pop	dword[.y4+12]


   push dword[esi+8]	      ;shuffle z
   pop	dword[.z1]
   push dword[esi+12+8]
   pop	dword[.z1+4]
   push dword[esi+24+8]
   pop	dword[.z1+8]
   push dword[esi+36+8]
   pop	dword[.z1+12]

   push dword[esi+36+12+8]
   pop	dword[.z2]
   push dword[esi+36+12*2+8]
   pop	dword[.z2+4]
   push dword[esi+36+12*3+8]
   pop	dword[.z2+8]
   push dword[esi+36+12*4+8]
   pop	dword[.z2+12]

   push dword[eax+4]
   pop	dword[.z3]
   push dword[eax+12+4]
   pop	dword[.z3+4]
   push dword[eax+12*2+4]
   pop	dword[.z3+8]
   push dword[eax+12*3+4]
   pop	dword[.z3+12]

   push dword[eax+12*4+4]
   pop	dword[.z4]
   push dword[eax+12*5+4]
   pop	dword[.z4+4]
   push dword[eax+12*6+4]
   pop	dword[.z4+8]
   push dword[eax+12*7+4]
   pop	dword[.z4+12]


;   mov  esp,[esp_repo]  ;.oldEsp



   movzx    ecx,dx
   mov	    .cs,ecx
 ;  sub      .cs,1
   cvtsi2ss xmm1,ecx
   rcpss    xmm1,xmm1
   movss    .ds,xmm1
   shr	    edx,16
   mov	    .ct,edx
;   sub      .ct,1
   cvtsi2ss xmm0,edx
   rcpss    xmm0,xmm0
   movss    .dt,xmm0

 ;  push     edi
   push     ebx

   xor	    ecx,ecx
 ;  mov      .s,ecx
   mov	    .t,ecx
   mov	    .vert_countB,ecx
   mov	    .tri_countB,ecx

 .looptt:
   push   ecx
   movss  xmm1,[f1]
   subss  xmm1,.t
   movss  xmm2,xmm1
   movss  xmm3,xmm1
   movss  xmm4,.t
   mulss  xmm2,xmm1
   mulss  xmm1,[f3]
   mulss  xmm3,xmm2
   movss  .a,xmm3
   mulss  xmm2,[f3]
   mulss  xmm4,xmm4
   mulss  xmm2,.t
   movss  .b,xmm2
   mulss  xmm1,xmm4
   movss  .c,xmm1
   mulss  xmm4,.t
   movss  .d,xmm4

   xor	    ecx,ecx
   mov	    dword .s,ecx

 .loopss:
   push  ecx
   movss xmm0,[f1]
   subss xmm0,.s
   movss xmm3,.s
   movss xmm1,xmm0
   mulss xmm1,xmm3
   mulss xmm1,[f3]
   movss xmm2,xmm1
   mulss xmm1,xmm0  ; xmm1 - bs
   mulss xmm2,xmm3  ; xmm2 - cs
   movss xmm4,xmm0
   mulss xmm0,xmm0
   mulss xmm0,xmm4  ; xmm0 - as
   mulss xmm3,xmm3
   mulss xmm3,.s    ; xmm3 - ds

   movaps   xmm4,.a
   movaps   xmm5,xmm4
   movaps   xmm6,xmm4
   movaps   xmm7,xmm4
   mulps    xmm4,[.x1]
   mulps    xmm5,[.x2]
   mulps    xmm6,[.x3]
   mulps    xmm7,[.x4]
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   mulss    xmm4,xmm0
   mulss    xmm5,xmm1
   mulss    xmm6,xmm2
   mulss    xmm7,xmm3
   addss    xmm4,xmm5
   addss    xmm6,xmm7
   addss    xmm4,xmm6
   movss    [edi],xmm4


   movaps   xmm4,.a
   movaps   xmm5,xmm4
   movaps   xmm6,xmm4
   movaps   xmm7,xmm4
   mulps    xmm4,[.y1]
   mulps    xmm5,[.y2]
   mulps    xmm6,[.y3]
   mulps    xmm7,[.y4]
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   mulss    xmm4,xmm0
   mulss    xmm5,xmm1
   mulss    xmm6,xmm2
   mulss    xmm7,xmm3
   addss    xmm4,xmm5
   addss    xmm6,xmm7
   addss    xmm4,xmm6
   movss    [edi+4],xmm4

   movaps   xmm4,.a
   movaps   xmm5,xmm4
   movaps   xmm6,xmm4
   movaps   xmm7,xmm4
   mulps    xmm4,[.z1]
   mulps    xmm5,[.z2]
   mulps    xmm6,[.z3]
   mulps    xmm7,[.z4]
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   haddps   xmm4,xmm4
   haddps   xmm5,xmm5
   haddps   xmm6,xmm6
   haddps   xmm7,xmm7
   mulss    xmm4,xmm0
   mulss    xmm5,xmm1
   mulss    xmm6,xmm2
   mulss    xmm7,xmm3
   addss    xmm4,xmm5
   addss    xmm6,xmm7
   addss    xmm4,xmm6
   movss    [edi+8],xmm4


   movss    xmm0,.s
   addss    xmm0,.ds
   movss    .s,xmm0
   add	    edi,12
   inc	    .vert_countB
   pop	    ecx
   inc	    ecx
   cmp	    ecx,.cs
   jna	    .loopss

   movss    xmm1,.t
   addss    xmm1,.dt
   movss    .t,xmm1
   pop	    ecx
   inc	    ecx
;   add      ecx,1
   cmp	    ecx,.ct
   jna	    .looptt

   ; vertices described one bezier surface calculated

   cmp	  ebx,-1
   cmove  edx,ebx
   je	  .end44
  ; calc triangles list
   cld
   mov	  edx,.ct
   mov	  esi,ebx
   xor	  ebx,ebx
 .oop1:
   mov	  ecx,.cs
  @@:
   mov	  eax,ebx

   mov	  [esi],eax
   inc	  eax
   mov	  [esi+20],eax
   mov	  [esi+8],eax
   add	  eax,.cs
   mov	  [esi+4],eax
   mov	  [esi+12],eax
   inc	  eax
   mov	  [esi+16],eax
   add	  esi,24
   inc	  ebx
   add	  .tri_countB,2
   loop   @b
   inc	  ebx
   dec	  edx
   jnz	  .oop1

;   pop      ebx
;   pop      edi

   mov	     ebx,.tri_countB
   imul      ebx,12
   pop	     ecx
;   add       ecx,ecx
   add	     ebx,ecx
  ; ebx - end of computed triangles list ptr
  ; edi - end of calculated vertices ptr

;   mov      ecx,.vert_countB
   mov	    edx,.tri_countB
 .end44:
   mov	    ecx,.vert_countB

   add	    esp,290
   pop	    ebp


ret
;==========================================================
draw_bezier_derives1:	   ; bezier surface
  ; esi - source  derives nodes
  ; ebx - souurce derv indexes (rectangles)
  ; edi - screen buffer
;  .screen equ [ebp-4]
  ; uses some globals
     push     ebp
     mov      ebp,esp
  ;   sub      esp,16
  ;   .counter equ dword[ebp-4]
     emms
;     mov      .screen,edi
     movzx    eax,[xres_var]
     movd     mm2,eax
   ;  mov      [screen_width_var],eax
   ;  mov      dword[lbuffer_ptr],edi
     movd     mm1,edi  ;.screen
     mov      eax,0x0000ff00
     movd     mm3,eax

  ;   mov      .counter,0
     movd     xmm7,dword[xres_var]
     mov      eax,1
     movd     xmm5,eax
     shufps   xmm5,xmm5,0
     pxor     xmm6,xmm6
     punpcklwd xmm7,xmm6
     movlhps  xmm7,xmm7
     psubd    xmm7,xmm5
;     push     esi
     push     ebx
     mov      ecx,4
  .again1:
     push     ebx
     push     ecx
     mov      ecx,3
  .again:
     push     ecx
     push     ebx
     movzx    eax,word[ebx]  ;.counter
     movzx    edx,word[ebx+2] ; .counter
     imul     edx,12
     add      edx,esi
     imul     eax,12
     add      eax,esi
     movlps   xmm0,[eax]
     movhps   xmm0,[edx]
 ;    movaps   xmm3,xmm0
     cvtps2dq xmm0,xmm0
     movaps   xmm2,xmm0
     movaps   xmm1,xmm0
     pcmpgtd  xmm0,xmm7
     pcmpgtd  xmm1,xmm6
     pxor     xmm0,xmm1
     pmovmskb eax,xmm0
     and      ah,al
     cmp      ah,0xff
     jne      .skip1


     ; eax=x1 ebx=y1 ecx=x2 edx=y2
     sub      esp,16
     movups   [esp],xmm2
     pop      eax ebx ecx edx

;     movd     eax,xmm2
     pxor     xmm6,xmm6
;     psrldq   xmm2,4
;     movd     ebx,xmm2
;     psrldq   xmm2,4
;     movd     ecx,xmm2
;     psrldq   xmm2,4
;     movd     edx,xmm2
     movd     xmm7,[xres_vard]
     punpcklwd xmm7,xmm6

     pushad


     call     linea1
     popad

 ;    movhlps  xmm0,xmm3
 ;    movzx    edx,[yres_var]
 ;    movzx    edi,[yres_var]
 ;    sub      edx,4
 ;    sub      edi,4
 ;    push     esi
 ;    push     edx
 ;    push     edi
 ;    call     bar_edge
 ;    pop      edi
 ;    pop      edx
 ;    pop      esi

   .skip1:
     pop     ebx
     add     ebx,2
     pop     ecx
   ;  dec      ecx
   ;  jnz     .again
     loop    .again

     pop     ecx
     pop     ebx
     add     ebx,8
     loop    .again1
 ;    sub     ecx,1
 ;    jnz     .again1

     pop     ebx

     mov      ecx,4
  .again2:
     push     ecx
     mov      ecx,3
  .again3:
     push     ecx
     push     ebx
     movzx    eax,word[ebx]
     movzx    edx,word[ebx+8]
     imul     eax,12
     imul     edx,12
     movlps   xmm0,[esi+eax]
     movhps   xmm0,[esi+edx]
     cvtps2dq xmm0,xmm0
     movaps   xmm2,xmm0
     movaps   xmm1,xmm0
     pcmpgtd  xmm0,xmm7
     pcmpgtd  xmm1,xmm6
     pxor     xmm0,xmm1
     pmovmskb eax,xmm0
     and      ah,al
     cmp      ah,0xff
     jne      .skip2

     ; eax=x1 ebx=y1 ecx=x2 edx=y2
     pxor     xmm6,xmm6
     movd     eax,xmm2
     psrldq   xmm2,4
     movd     ebx,xmm2
     psrldq   xmm2,4
     movd     ecx,xmm2
     psrldq   xmm2,4
     movd     edx,xmm2
     movd     xmm7,[xres_vard]
     punpcklwd xmm7,xmm6

     pushad
     call     linea1
     popad
   .skip2:
     pop     ebx
     add     ebx,2
     pop     ecx
     dec     ecx
     jnz     .again3
     pop     ecx
     dec     ecx
     jnz     .again2

     mov   esp,ebp
     pop   ebp

ret

draw_bezier_bars:
;ret
; in:
;  esi - ptr to derives vertices list
;  ebx - ptr to derives indexes (16 pro one rectangle)
;  ecx - number of surfaces
;  globals - screen_ptr, i12, ..
   push   ebp
   mov	  ebp,esp
   .ym4   equ dword[ebp-4]

   shl	    ecx,4
   movzx    edx,[xres_var]
   movzx    edi,[yres_var]
   sub	    edx,4
   sub	    edi,4
   push     edi

   pxor     xmm0,xmm0
 .oop:
   push   ecx
   movzx  eax,word[ebx]
   movzx  ecx,word[ebx]
   imul   eax,12
   add	  eax,esi
   movlps xmm0,[eax]
   push   ebx
   push   edx
   push   esi
   mov	  esi,.ym4
   mov	  eax,0x00ffff00
 ;  ecx - edge index; signature
 ;  xmm0 - coordinates
 ;  eax - color
 ;  edx  - xresm4
 ;  esi  - yresm4
   call   bar_edge ; edit.inc
   pop	  esi
   pop	  edx
   pop	  ebx
   add	  ebx,2
   pop	  ecx
   loop   .oop

   mov	  esp,ebp
   pop	  ebp
ret
